/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <Brother_HL_10V_Resolutions.hpp>
#include <JobProperties.hpp>

Brother_HL_10V_Resolutions::
Brother_HL_10V_Resolutions (Device *pDevice,
                          PSZCRO      pszJobProperties,
                          int         iXInternalRes,
                          int         iYInternalRes,
                          BinaryData *data,
                          int         iCapabilities,
                          int         iDestinationBitsPerPel,
                          int         iScanlineMultiple)
      : DeviceResolution (pDevice,
                          pszJobProperties,
                          iXInternalRes,
                          iYInternalRes,
                          data,
                          iCapabilities,
                          iDestinationBitsPerPel,
                          iScanlineMultiple)
{
}

static byte abData100x100[] = {27,42,116,49,48,48,82};
static byte abData150x150[] = {27,42,116,49,53,48,82};
static byte abData300x300[] = {27,42,116,51,48,48,82};
static byte abData75x75[] = {27,42,116,55,53,82};

typedef struct _MappingTable {
   PSZCRO pszName;
   PSZCRO pszJobProperties;
   PSZCRO pszJobPropertiesDevice;
   int    iXInternalRes;
   int    iYInternalRes;
   byte  *pbBinaryData;
   int    cbBinaryData;
   int    iCapabilities;
   int    iDestinationBitsPerPel;
   int    iScanLineMultiple;
} MAPPINGTABLE, *PMAPPINGTABLE, **PPMAPPINGTABLE;

static MAPPINGTABLE aTable[] = {
   { "100x100", "Resolution=100x100", 0, 0, 0, abData100x100, 7, 0, 1, 1 },
   { "150x150", "Resolution=150x150", 0, 0, 0, abData150x150, 7, 0, 1, 1 },
   { "300x300", "Resolution=300x300", 0, 0, 0, abData300x300, 7, 0, 1, 1 },
   { "75x75", "Resolution=75x75", 0, 0, 0, abData75x75, 6, 0, 1, 1 }
};

static int findTableEntry (PSZCRO pszResolutionName);

int
findTableEntry (PSZCRO pszResolutionName)
{
   int   iLow     = 0;
   int   iMid     = (int)dimof (aTable) / 2;
   int   iHigh    = (int)dimof (aTable) - 1;
   int   iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszResolutionName, aTable[iMid].pszName);

      if (0 == iResult)
      {
         return iMid;
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return -1;
}

DeviceResolution * Brother_HL_10V_Resolutions::
createS (Device *pDevice, PSZCRO pszJobProperties)
{
   JobProperties          jobProp (pszJobProperties);
   JobPropertyEnumerator *pEnum                      = 0;

   pEnum = jobProp.getEnumeration ();

   while (pEnum->hasMoreElements ())
   {
      PSZCRO pszKey   = pEnum->getCurrentKey ();
      PSZCRO pszValue = pEnum->getCurrentValue ();

      if (0 == strcmp (pszKey, "Resolution"))
      {
         int iIndex = -1;

         iIndex = findTableEntry (pszValue);

         if (0 <= iIndex)
         {
            return new Brother_HL_10V_Resolutions (pDevice, aTable[iIndex].pszJobProperties, aTable[iIndex].iXInternalRes, aTable[iIndex].iYInternalRes, new BinaryData (aTable[iIndex].pbBinaryData, aTable[iIndex].cbBinaryData), aTable[iIndex].iCapabilities, aTable[iIndex].iDestinationBitsPerPel, aTable[iIndex].iScanLineMultiple);
         }
      }

      pEnum->nextElement ();
   }

   delete pEnum;

   return 0;
}

DeviceResolution * Brother_HL_10V_Resolutions::
create (Device *pDevice, PSZCRO pszJobProperties)
{
   return createS (pDevice, pszJobProperties);
}

bool Brother_HL_10V_Resolutions ::
isSupported (PSZCRO pszJobProperties)
{
   JobProperties          jobProp (pszJobProperties);
   JobPropertyEnumerator *pEnum                      = 0;
   bool                   fRet                       = false;

   pEnum = jobProp.getEnumeration ();

   while (pEnum->hasMoreElements ())
   {
      PSZCRO pszKey   = pEnum->getCurrentKey ();
      PSZCRO pszValue = pEnum->getCurrentValue ();

      if (0 == strcmp (pszKey, "Resolution"))
      {
         fRet = 0 <= findTableEntry (pszValue);
      }

      pEnum->nextElement ();
   }

   delete pEnum;

   return fRet;
}

Enumeration * Brother_HL_10V_Resolutions::
getEnumeration (bool fInDeviceSpecific)
{
   class ResolutionEnumerator : public Enumeration
   {
   public:
      ResolutionEnumerator (Device *pDevice, int cResolutions, PMAPPINGTABLE aTable, bool fInDeviceSpecific)
      {
         pDevice_d           = pDevice;
         iResolution_d       = 0;
         cResolutions_d      = cResolutions;
         aTable_d            = aTable;
         fInDeviceSpecific_d = fInDeviceSpecific;
      }

      virtual bool hasMoreElements ()
      {
         if (iResolution_d < cResolutions_d)
            return true;
         else
            return false;
      }

      virtual void *nextElement ()
      {
         void *pvRet = 0;

         if (iResolution_d > cResolutions_d - 1)
            return 0;

         if (fInDeviceSpecific_d)
         {
            if (aTable_d[iResolution_d].pszJobPropertiesDevice)
            {
               pvRet = (void *)new JobProperties (aTable_d[iResolution_d].pszJobPropertiesDevice);
            }
         }

         if (!pvRet)
         {
            pvRet = (void *)new JobProperties (aTable_d[iResolution_d].pszJobProperties);
         }

         iResolution_d++;

         return pvRet;
      }

   private:
      Device       *pDevice_d;
      int           iResolution_d;
      int           cResolutions_d;
      PMAPPINGTABLE aTable_d;
      bool          fInDeviceSpecific_d;
   };

   return new ResolutionEnumerator (pDevice_d, dimof (aTable), aTable, fInDeviceSpecific);
}
