/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <Kyocera_Orientations.hpp>
#include <JobProperties.hpp>

Kyocera_Orientations::
Kyocera_Orientations (Device *pDevice,
                     PSZCRO  pszJobProperties)
      : DeviceOrientation (pDevice,
                           pszJobProperties)
{
}

typedef struct _MappingTable {
   PSZCRO pszName;
   PSZCRO pszJobProperties;
   PSZCRO pszJobPropertiesDevice;
} MAPPINGTABLE, *PMAPPINGTABLE, **PPMAPPINGTABLE;

static MAPPINGTABLE aTable[] = {
   { "Portrait", "Rotation=Portrait", 0 }
};

static int findTableEntry (PSZCRO pszOrientationName);

int
findTableEntry (PSZCRO pszOrientationName)
{
   int   iLow     = 0;
   int   iMid     = (int)dimof (aTable) / 2;
   int   iHigh    = (int)dimof (aTable) - 1;
   int   iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszOrientationName, aTable[iMid].pszName);

      if (0 == iResult)
      {
         return iMid;
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return -1;
}

DeviceOrientation * Kyocera_Orientations::
createS (Device *pDevice, PSZCRO pszJobProperties)
{
   JobProperties          jobProp (pszJobProperties);
   JobPropertyEnumerator *pEnum                      = 0;

   pEnum = jobProp.getEnumeration ();

   while (pEnum->hasMoreElements ())
   {
      PSZCRO pszKey   = pEnum->getCurrentKey ();
      PSZCRO pszValue = pEnum->getCurrentValue ();

      if (0 == strcmp (pszKey, "Rotation"))
      {
         int iIndex = -1;

         iIndex = findTableEntry (pszValue);

         if (0 <= iIndex)
         {
            return new Kyocera_Orientations (pDevice, aTable[iIndex].pszJobProperties);
         }
      }

      pEnum->nextElement ();
   }

   delete pEnum;

   return 0;
}

DeviceOrientation * Kyocera_Orientations::
create (Device *pDevice, PSZCRO pszJobProperties)
{
   return createS (pDevice, pszJobProperties);
}

bool Kyocera_Orientations ::
isSupported (PSZCRO pszJobProperties)
{
   JobProperties          jobProp (pszJobProperties);
   JobPropertyEnumerator *pEnum                      = 0;
   bool                   fRet                       = false;

   pEnum = jobProp.getEnumeration ();

   while (pEnum->hasMoreElements ())
   {
      PSZCRO pszKey   = pEnum->getCurrentKey ();
      PSZCRO pszValue = pEnum->getCurrentValue ();

      if (0 == strcmp (pszKey, "Rotation"))
      {
         fRet = 0 <= findTableEntry (pszValue);
      }

      pEnum->nextElement ();
   }

   delete pEnum;

   return fRet;
}

Enumeration * Kyocera_Orientations::
getEnumeration (bool fInDeviceSpecific)
{
   class OrientationEnumerator : public Enumeration
   {
   public:
      OrientationEnumerator (Device *pDevice, int cOrientations, PMAPPINGTABLE aTable, bool fInDeviceSpecific)
      {
         pDevice_d           = pDevice;
         iOrientation_d      = 0;
         cOrientations_d     = cOrientations;
         aTable_d            = aTable;
         fInDeviceSpecific_d = fInDeviceSpecific;
      }

      virtual bool hasMoreElements ()
      {
         if (iOrientation_d < cOrientations_d)
            return true;
         else
            return false;
      }

      virtual void *nextElement ()
      {
         void *pvRet = 0;

         if (iOrientation_d > cOrientations_d - 1)
            return pvRet;

         if (fInDeviceSpecific_d)
         {
            if (aTable_d[iOrientation_d].pszJobPropertiesDevice)
            {
               pvRet = (void *)new JobProperties (aTable_d[iOrientation_d].pszJobPropertiesDevice);
            }
         }

         if (!pvRet)
         {
            pvRet = (void *)new JobProperties (aTable_d[iOrientation_d].pszJobProperties);
         }

         iOrientation_d++;

         return pvRet;
      }

   private:
      Device       *pDevice_d;
      int           iOrientation_d;
      int           cOrientations_d;
      PMAPPINGTABLE aTable_d;
      bool          fInDeviceSpecific_d;
   };

   return new OrientationEnumerator (pDevice_d, dimof (aTable), aTable, fInDeviceSpecific);
}
