/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include <Panasonic_KX_P1124_PrintModes.hpp>
#include <JobProperties.hpp>

Panasonic_KX_P1124_PrintModes::
Panasonic_KX_P1124_PrintModes (Device *pDevice,
                          PSZCRO      pszJobProperties,
                          int         iPhysicalCount,
                          int         iLogicalCount,
                          int         iPlanes)
      : DevicePrintMode (pDevice,
                         pszJobProperties,
                         iPhysicalCount,
                         iLogicalCount,
                         iPlanes)
{
}

typedef struct _MappingTable {
   PSZCRO pszName;
   PSZCRO pszJobProperties;
   PSZCRO pszJobPropertiesDevice;
   int    iPhysicalCount;
   int    iLogicalCount;
   int    iPlanes;
} MAPPINGTABLE, *PMAPPINGTABLE, **PPMAPPINGTABLE;

static MAPPINGTABLE aTable[] = {
   { "PRINT_MODE_1_ANY", "printmode=PRINT_MODE_1_ANY", 0, 1, 1, 1 },
   { "PRINT_MODE_24_CMYK", "printmode=PRINT_MODE_24_CMYK", 0, 4, 24, 1 },
   { "PRINT_MODE_8_CMYK", "printmode=PRINT_MODE_8_CMYK", 0, 4, 8, 1 }
};

static int findTableEntry (PSZCRO pszPrintModeName);

int
findTableEntry (PSZCRO pszPrintModeName)
{
   int   iLow     = 0;
   int   iMid     = (int)dimof (aTable) / 2;
   int   iHigh    = (int)dimof (aTable) - 1;
   int   iResult;

   while (iLow <= iHigh)
   {
      iResult = strcmp (pszPrintModeName, aTable[iMid].pszName);

      if (0 == iResult)
      {
         return iMid;
      }
      else if (0 > iResult)
      {
         // str1 < str2
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (0 < iResult)
      {
         // str1 > str2
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   return -1;
}

DevicePrintMode * Panasonic_KX_P1124_PrintModes::
createS (Device *pDevice, PSZCRO pszJobProperties)
{
   JobProperties          jobProp (pszJobProperties);
   JobPropertyEnumerator *pEnum                      = 0;

   pEnum = jobProp.getEnumeration ();

   while (pEnum->hasMoreElements ())
   {
      PSZCRO pszKey   = pEnum->getCurrentKey ();
      PSZCRO pszValue = pEnum->getCurrentValue ();

      if (0 == strcmp (pszKey, "printmode"))
      {
         int iIndex = -1;

         iIndex = findTableEntry (pszValue);

         if (0 <= iIndex)
         {
            return new Panasonic_KX_P1124_PrintModes (pDevice, aTable[iIndex].pszJobProperties, aTable[iIndex].iPhysicalCount, aTable[iIndex].iLogicalCount, aTable[iIndex].iPlanes);
         }
      }

      pEnum->nextElement ();
   }

   delete pEnum;

   return 0;
}

DevicePrintMode * Panasonic_KX_P1124_PrintModes::
create (Device *pDevice, PSZCRO pszJobProperties)
{
   return createS (pDevice, pszJobProperties);
}

bool Panasonic_KX_P1124_PrintModes ::
isSupported (PSZCRO pszJobProperties)
{
   JobProperties          jobProp (pszJobProperties);
   JobPropertyEnumerator *pEnum                      = 0;
   bool                   fRet                       = false;

   pEnum = jobProp.getEnumeration ();

   while (pEnum->hasMoreElements ())
   {
      PSZCRO pszKey   = pEnum->getCurrentKey ();
      PSZCRO pszValue = pEnum->getCurrentValue ();

      if (0 == strcmp (pszKey, "printmode"))
      {
         fRet = 0 <= findTableEntry (pszValue);
      }

      pEnum->nextElement ();
   }

   delete pEnum;

   return fRet;
}

Enumeration * Panasonic_KX_P1124_PrintModes::
getEnumeration (bool fInDeviceSpecific)
{
   class PrintModeEnumerator : public Enumeration
   {
   public:
      PrintModeEnumerator (Device *pDevice, int cPrintModes, PMAPPINGTABLE aTable, bool fInDeviceSpecific)
      {
         pDevice_d           = pDevice;
         iPrintMode_d        = 0;
         cPrintModes_d       = cPrintModes;
         aTable_d            = aTable;
         fInDeviceSpecific_d = fInDeviceSpecific;
      }

      virtual bool hasMoreElements ()
      {
         if (iPrintMode_d < cPrintModes_d)
            return true;
         else
            return false;
      }

      virtual void *nextElement ()
      {
         void *pvRet = 0;

         if (iPrintMode_d > cPrintModes_d - 1)
            return 0;

         if (fInDeviceSpecific_d)
         {
            if (aTable_d[iPrintMode_d].pszJobPropertiesDevice)
            {
               pvRet = (void *)new JobProperties (aTable_d[iPrintMode_d].pszJobPropertiesDevice);
            }
         }

         if (!pvRet)
         {
            pvRet = (void *)new JobProperties (aTable_d[iPrintMode_d].pszJobProperties);
         }

         iPrintMode_d++;

         return pvRet;
      }

   private:
      Device       *pDevice_d;
      int           iPrintMode_d;
      int           cPrintModes_d;
      PMAPPINGTABLE aTable_d;
      bool          fInDeviceSpecific_d;
   };

   return new PrintModeEnumerator (pDevice_d, dimof (aTable), aTable, fInDeviceSpecific);
}
